/*
	Example program for DVD-R access on Wii in GameCube mode.
    Copyright (C) 2008 WiiGator.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <malloc.h>
#include <ogcsys.h>
#include <gccore.h>

/* Error codes */
#define ERR_EINVAL		 -1
#define ERR_ENOENT		 -6
#define ERR_EIO			  2
#define ERR_EINCMD		128

/* Constants */
#define SECTOR_SIZE		0x800

/* DVD drive */
volatile long *dvdio  = (volatile long *)0xCC006000;
unsigned char dvdbuf[SECTOR_SIZE] __attribute__((aligned(32)));

/* Variables */
unsigned long offset_base = 0;
static char hexmap[] = {
    '0', '1', '2', '3', '4', '5', '6', '7',
    '8', '9', 'a', 'b', 'c', 'd', 'e', 'f' 
};

void debug_printf(const char *fmt, ...) {
  char buf[1024];
  int len;
  va_list ap;
  usb_flush(1);
  va_start(ap, fmt);
  len = vsnprintf(buf, sizeof(buf), fmt, ap);
  vprintf(fmt, ap);
  va_end(ap);
  if (len <= 0 || len > sizeof(buf)) printf("Error: len = %d\n", len);
  else usb_sendbuffer(1, buf, len);
}


void hexDump(unsigned char *buffer, unsigned long size)
{
	int i;

	for (i = 0; i < size; i++) {
		if ((i > 0) && ((i & 0xF) == 0)) {
			debug_printf("\n");
		}
		unsigned char c;

		c = buffer[i];
		debug_printf("%c", hexmap[(c >> 4) & 0xf]);
		debug_printf("%c", hexmap[c & 0xf]);
		if ((i + 1) != size) {
			debug_printf(" ");
		}
	}
	debug_printf("\n");
}

int DI_Inquiry(void *outbuf)
{
	/* Reset notify */
	dvdio[0] = 0x2E;
	dvdio[1] = 0;
	dvdio[2] = 0x12000000;
	dvdio[3] = 0;
	dvdio[4] = 0x20;
	dvdio[5] = (unsigned long)outbuf;
	dvdio[6] = 0x20;
	dvdio[7] = 3;

	while (dvdio[7] & 1);

	/* Invalidate cache */
	DCInvalidateRange(outbuf, 0x20);

	return 1;
}

unsigned int DI_GetError(void)
{
	/* Get last error. */
	dvdio[2] = 0xE0000000;
	dvdio[7] = 1;

	while (dvdio[7] & 1);

	return dvdio[8];
}

int DI_ReadDvd(void *outbuf, unsigned long len, unsigned long lba)
{
	/* Disc read */
	dvdio[0] = 0;
	dvdio[1] = 0;
	dvdio[2] = 0xD0000000;
	dvdio[3] = lba;
	dvdio[4] = len >> 11;
	dvdio[5] = (unsigned long)outbuf;
	dvdio[6] = len;
	dvdio[7] = 3;

	while (dvdio[7] & 1);

	/* Invalidate cache */
	DCInvalidateRange(outbuf, len);

	if (!(dvdio[0] & 0x4))
		return 1;

	return ERR_EIO;
}

int DI_ReadDvdA8DiscId(void *outbuf)
{
	dvdio[0] = 0;
	dvdio[1] = 0;
	dvdio[2] = 0xA8000040;
	dvdio[3] = 0;
	dvdio[4] = 0x20;
	dvdio[5] = (unsigned long)outbuf;
	dvdio[6] = 0x20;
	dvdio[7] = 3;

	while (dvdio[7] & 1);

	/* Invalidate cache */
	DCInvalidateRange(outbuf, 0x20);

	if (!(dvdio[0] & 0x4))
		return 1;

	return ERR_EIO;
}

int DI_ReadDvdA8(void *outbuf, unsigned long len, unsigned long offset)
{
	/* Disc read */
	dvdio[0] = 0;
	dvdio[1] = 0;
	dvdio[2] = 0xA8000000;
	dvdio[3] = offset;
	dvdio[4] = len;
	dvdio[5] = (unsigned long)outbuf;
	dvdio[6] = len;
	dvdio[7] = 3;

	while (dvdio[7] & 1);

	/* Invalidate cache */
	DCInvalidateRange(outbuf, len);

	if (!(dvdio[0] & 0x4))
		return 1;

	return ERR_EIO;
}

int DI_StopMotor(unsigned long eject, unsigned long kill)
{
	/* Stop motor */
	dvdio[0] = 0x2E;
	dvdio[1] = 0;
	dvdio[2] = 0xE3000000;
	dvdio[3] = eject;
	dvdio[4] = kill;
	dvdio[5] = 0;
	dvdio[6] = 0;
	dvdio[7] = 1;

	while (dvdio[7] & 1);

	if (dvdio[0] & 0x4)
		return 1;

	return ERR_EIO;
}

int DI_Offset(unsigned long offset)
{
	/* Set offset */
	dvdio[0] = 0x2E;
	dvdio[1] = 0;
	dvdio[2] = 0xD9000000;
	dvdio[3] = offset;
	dvdio[4] = 0;
	dvdio[5] = 0;
	dvdio[6] = 0;
	dvdio[7] = 1;

	while (dvdio[7] & 1);

	if (dvdio[0] & 0x4)
		return 1;

	return ERR_EIO;
}

int DI_GetStatus(void *outbuf)
{
	/* Get cover state */
	dvdio[0] = 0;
	dvdio[1] = 0;
	dvdio[2] = 0xE0000000;
	dvdio[3] = 0;
	dvdio[4] = 0;
	dvdio[5] = 0;
	dvdio[6] = 0;
	dvdio[7] = 1;

	while (dvdio[7] & 1);

	/* Copy cover state */
	memcpy(outbuf, (void *)&dvdio[8], 4);

	return 1;
}

int dvdtest(void)
{
	unsigned long rv;
	unsigned long offset;
	int i;

	offset = 0;
	debug_printf("Gamecube DVD-R Test Program\n");
	debug_printf("Last error 0x%x\n", DI_GetError());
	if (!(dvdio[1] & 1)) {
#if 0
		debug_printf("Read Disc ID\n");
		memset(dvdbuf, 0, SECTOR_SIZE);
	 	DCFlushRange(dvdbuf, SECTOR_SIZE);
		DI_ReadDvdA8DiscId(dvdbuf);
		debug_printf("Last error 0x%x\n", DI_GetError());
		hexDump(dvdbuf, 0x20);
		debug_printf("\n");
#endif

		debug_printf("Read DVD\n");
		memset(dvdbuf, 0, SECTOR_SIZE);
	 	DCFlushRange(dvdbuf, SECTOR_SIZE);
		rv = DI_ReadDvd(dvdbuf, SECTOR_SIZE, offset >> 11);
		debug_printf("rv = %d, Last error 0x%x\n", rv, DI_GetError());
		debug_printf("DI_ReadDvd offset 0x%x\n", offset);
		hexDump(dvdbuf, 0x40);
	} else {
		debug_printf("Test failed no DVD inserted.\n");
	}
	return rv;
}
