
/*
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */
#include <stdint.h>
#include <stdio.h>
#include <unistd.h>
#include <gctypes.h>
#include <ogc/cache.h>
#include <ogcsys.h>
#include <gccore.h>
#include <string.h>

#include "dvd.h"
#include "anaconda.h"
#include "debugprintf.h"
#include "loadapp.h"
#include "loadmios.h"
#include "screen.h"

#define _CPU_ISR_Disable( _isr_cookie ) \
  { register u32 _disable_mask = 0; \
	_isr_cookie = 0; \
    __asm__ __volatile__ ( \
	  "mfmsr %0\n" \
	  "rlwinm %1,%0,0,17,15\n" \
	  "mtmsr %1\n" \
	  "extrwi %0,%0,1,16" \
	  : "=&r" ((_isr_cookie)), "=&r" ((_disable_mask)) \
	  : "0" ((_isr_cookie)), "1" ((_disable_mask)) \
	); \
  }
#define SECTOR_SIZE 0x800

#define DVD_CHECK() \
	if (rv != 0) { \
		printf("   Failed to read DVD %d.\n", __LINE__); \
		sleep(5); \
		call_mios(); \
	}

#define MAX_GAMES 64

extern void __exception_closeall();
static char dvdid[6];
unsigned int gameOffsetTable[MAX_GAMES];


#define MEM_VIDEO_BASE  			 (0xCC002000)
#define GC_INIT_BASE			   	 (0x80000020)
typedef struct {
	u8 ConsoleID;				//G = Gamecube, R = Wii.
	u8 GamecodeA;				//2 Ascii letters to indicate the GameID.
	u8 GamecodeB;				//2 Ascii letters to indicate the GameID.
	u8 CountryCode;				//J=JAP . P=PAL . E=USA . D=OoT MasterQuest
	u8 MakerCodeA;				//Eg 08 = Sega etc.
	u8 MakerCodeB;
	u8 DiscID;
	u8 Version;
	u8 AudioStreaming;			//01 = Enable it. 00 = Don't
	u8 StreamBufSize;			//For the AudioEnable. (always 00?)
	u8 unused_1[18];
	u32 DVDMagicWord;			//0xC2339F3D
	char GameName[992];			//String Name of Game, rarely > 32 chars
	u32 DMonitorOffset;			//offset of debug monitor (dh.bin)?
	u32 DMonitorLoadAd;			//addr(?) to load debug monitor?
	u8 unused_2[24];
	u32 DOLOffset;				//offset of main executable DOL (bootfile)
	u32 FSTOffset;				//offset of the FST ("fst.bin")
	u32 FSTSize;				//size of FST
	u32 MaxFSTSize;				//maximum size of FST (usually same as FSTSize)*
	u32 UserPos;				//user position(?)
	u32 UserLength;				//user length(?)
	u32 unknown;				//(?)
	u32 unused_3;
} DVDINFO;

static DVDINFO GCMDisk __attribute((aligned(32)));

/*
Region Defaults - tmbinc
*/
void Set_Game_Videomode(int mode)
{
	static const u32 vi_Mode640X480NtscYUV16[32] =
		{ 0x0F060001, 0x476901AD, 0x02EA5140, 0x00030018, 0x00020019,
		0x410C410C, 0x40ED40ED, 0x00435A4E, 0x00000000, 0x00435A4E,
		0x00000000, 0x00000000, 0x110701AE, 0x10010001, 0x00010001,
		0x00010001, 0x00000000, 0x00000000, 0x28500100, 0x1AE771F0,
		0x0DB4A574, 0x00C1188E, 0xC4C0CBE2, 0xFCECDECF, 0x13130F08,
		0x00080C0F, 0x00FF0000, 0x00000000, 0x02800000, 0x000000FF,
		0x00FF00FF, 0x00FF00FF
	};
	static const u32 vi_Mode640x480NtscProgressiveYUV16[32] =
		{ 0x1e0c0005, 0x476901ad, 0x02ea5140, 0x00060030, 0x00060030,
		0x81d881d8, 0x81d881d8, 0x10000000, 0x00000000, 0x00000000,
		0x00000000, 0x037702b6, 0x90010001, 0x00000000, 0x00000000,
		0x00000000, 0x00000000, 0x00000000, 0x28280100, 0x1ae771f0,
		0x0db4a574, 0x00c1188e, 0xc4c0cbe2, 0xfcecdecf, 0x13130f08,
		0x00080c0f, 0x00ff0000, 0x00010001, 0x02800000, 0x000000ff,
		0x00ff00ff, 0x00ff00ff
	};
	static const u32 vi_Mode640X576Pal50YUV16[32] =
		{ 0x11F50101, 0x4B6A01B0, 0x02F85640, 0x00010023, 0x00000024,
		0x4D2B4D6D, 0x4D8A4D4C, 0x0066D480, 0x00000000, 0x0066D980,
		0x00000000, 0x00C901F3, 0x913901B1, 0x90010001, 0x00010001,
		0x00010001, 0x00000000, 0x00000000, 0x28500100, 0x1AE771F0,
		0x0DB4A574, 0x00C1188E, 0xC4C0CBE2, 0xFCECDECF, 0x13130F08,
		0x00080C0F, 0x00FF0000, 0x00000000, 0x02800000, 0x000000FF,
		0x00FF00FF, 0x00FF00FF
	};
	static const u32 vi_Mode640X480Pal60YUV16[32] =
		{ 0x0F060001, 0x476901AD, 0x02EA5140, 0x00030018, 0x00020019,
		0x410C410C, 0x40ED40ED, 0x0066D480, 0x00000000, 0x0066D980,
		0x00000000, 0x00C9010F, 0x910701AE, 0x90010001, 0x00010001,
		0x00010001, 0x00000000, 0x00000000, 0x28500100, 0x1AE771F0,
		0x0DB4A574, 0x00C1188E, 0xC4C0CBE2, 0xFCECDECF, 0x13130F08,
		0x00080C0F, 0x00FF0000, 0x00000000, 0x02800000, 0x000000FF,
		0x00FF00FF, 0x00FF00FF
	};
	if (mode == 0)
		memcpy((void *) MEM_VIDEO_BASE, vi_Mode640X480NtscYUV16, 0x80);	//ntsc
	if (mode == 1)
		memcpy((void *) MEM_VIDEO_BASE, vi_Mode640X576Pal50YUV16, 0x80);	//pal
	if (mode == 2)
		memcpy((void *) MEM_VIDEO_BASE, vi_Mode640X480Pal60YUV16, 0x80);	//pal60
	if (mode == 3)
		memcpy((void *) MEM_VIDEO_BASE, vi_Mode640x480NtscProgressiveYUV16, 0x80);	//480ntsc
}

/*
Initialization routine with the defaults
*/
void InitSystem(unsigned long VidMode)
{
	unsigned long level = 0;
	static u32 GC_DefaultConfig[64] = { 0x0D15EA5E, 0x00000001, 0x01800000, 0x00000003,	//  0.. 3 80000020
		0x00000000, 0x816FFFF0, 0x817FE8C0, 0x00000024,	//  4.. 7 80000030
		0x00000000, 0x00000000, 0x00000000, 0x00000000,	//  8..11 80000040
		0x00000000, 0x00000000, 0x00000000, 0x00000000,	// 12..15 80000050
		0x38A00040, 0x7C6802A6, 0x9065000C, 0x80650008,	// 16..19 80000060
		0x64638000, 0x7C6803A6, 0x38600030, 0x7C600124,	// 20..23 80000070
		0x4E800020, 0x00000000, 0x00000000, 0x00000000,	// 24..27 80000080
		0x00000000, 0x00000000, 0x00000000, 0x00000000,	// 28..31 80000090
		0x00000000, 0x00000000, 0x00000000, 0x00000000,	// 32..35 800000A0
		0x00000000, 0x00000000, 0x00000000, 0x00000000,	// 36..39 800000B0
		0x015D47F8, 0xF8248360, 0x00000000, 0x00000001,	// 40..43 800000C0
		0x00000000, 0x00000000, 0x00000000, 0x00000000,	// 44..47 800000D0
		0x814B7F50, 0x815D47F8, 0x00000000, 0x81800000,	// 48..51 800000E0
		0x01800000, 0x817FC8C0, 0x09A7EC80, 0x1CF7C580	// 52..55 800000F0
	};

	//Lets close Exceptions
	_CPU_ISR_Disable(level);
	__exception_closeall();
	if ((VidMode == 0) || (VidMode == 0x45) || (VidMode == 0x4A)) {	//Standard NTSC
		Set_Game_Videomode(0);
		GC_DefaultConfig[43] = 0;
	}
	if ((VidMode == 1) || (VidMode == 0x50) || (VidMode == 0x55)) {	//Standard PAL50
		Set_Game_Videomode(1);
		GC_DefaultConfig[43] = 1;
	}
	if (VidMode == 2) {			//PAL60
		Set_Game_Videomode(2);
		GC_DefaultConfig[43] = 5;
	}
	if (VidMode == 3) {			//480P
		Set_Game_Videomode(3);
		GC_DefaultConfig[43] = 0;
	}
	memcpy((void *) GC_INIT_BASE, GC_DefaultConfig, 0xe0);
	DCFlushRange((void *) 0x80000000, 0x3100);

	/* Clear memory. */
	memset((void *) 0x80003100, 0, 0x00d00000 - 0x3100);
	DCFlushRange((void *) 0x80003100, 0x00d00000 - 0x3100);
} u32 tempOffset = 0;

void check_game(int activeGame, int count)
{
	char tempStr[32];
	int rv;

	rv = dvd_read(&GCMDisk, sizeof(GCMDisk), 0);
	DVD_CHECK();

	clearScreen();
	memset(tempStr, 0, 40);
	strncpy(tempStr, GCMDisk.GameName, 38);
	debug_printf("Game ID: 0x%c%c%c%c%c%c %s\n",
		GCMDisk.ConsoleID,
		GCMDisk.GamecodeA,
		GCMDisk.GamecodeB,
		GCMDisk.CountryCode,
		GCMDisk.MakerCodeA,
		GCMDisk.MakerCodeB,
		tempStr);

	if (count > 1) {
		printf("   [%d] %s\n\n", activeGame, tempStr);
	} else {
		printf("   %s\n\n", tempStr);
	}
	printf("   Press: X for PAL   L for PAL60\n");
	printf("          Y for NTSC  R for 480p\n");
	printf("          A for Auto Detect\n");
	printf("          B to use patched MIOS for start\n");
#ifdef RELOAD_SUPPORT
	printf("          START for Reloader (slow)\n");
#endif
	if (count > 1) {
		printf("          Left/Right to switch game\n");
	}
}

void setOffset(int activeGame)
{
	base_offset = gameOffsetTable[activeGame];
}

void FinalizeBoot(int count)
{
	volatile int activeGame = 0;

	setOffset(activeGame);
	check_game(activeGame, count);

	while (1) {
		PAD_ScanPads();
		int buttonsDown = PAD_ButtonsDown(0);

		if (buttonsDown & PAD_BUTTON_Y) {
			InitSystem(0);
			load_app(0);
			break;
		}
		if (buttonsDown & PAD_BUTTON_X) {
			InitSystem(1);
			load_app(0);
			break;
		}
		if (buttonsDown & PAD_TRIGGER_L) {
			InitSystem(2);
			load_app(0);
			break;
		}
		if (buttonsDown & PAD_TRIGGER_R) {
			InitSystem(3);
			load_app(0);
			break;
		}
		if (buttonsDown & PAD_BUTTON_A) {
			InitSystem(GCMDisk.CountryCode);
			load_app(0);
			break;
		}
#ifdef RELOAD_SUPPORT
		if (buttonsDown & PAD_BUTTON_START) {
			InitSystem(GCMDisk.CountryCode);
			load_app(1);
			break;
		}
#endif
		if (buttonsDown & PAD_BUTTON_B) {
			/* Start using MIOS with backup support. */
			load_mios();
			break;
		}
		if (count > 1) {
			if (buttonsDown & PAD_BUTTON_LEFT) {
				activeGame--;
				if (activeGame < 0) {
					activeGame++;
				}
				setOffset(activeGame);
				check_game(activeGame, count);
			}
			if (buttonsDown & PAD_BUTTON_RIGHT) {
				activeGame++;
				if (activeGame >= count) {
					activeGame--;
				}
				setOffset(activeGame);
				check_game(activeGame, count);
			}
		}
	};
}

void multiboot(void)
{
	int rv;
	int i;
	int count;
	uint32_t type;

	rv = dvd_read(gameOffsetTable, sizeof(gameOffsetTable), 64);
	DVD_CHECK();

	dvd_read(&type, sizeof(type),4);

	count = 0;
	for (i = 0; i < MAX_GAMES; i++) {
		if (gameOffsetTable[i] != 0) {
			if(type != 0x44564439) {
				/* Not DVD9 offsets need to convert the offset. */
				gameOffsetTable[i] = gameOffsetTable[i] >> 2;
			}
			count++;
		} else {
			break;
		}
	}

	FinalizeBoot(count);
}

void anaconda04(void)
{
	unsigned char checkBuffer[32];
	int retry = -1;
	int rv;

	while (retry) {
		clearScreen();
	
		base_offset = 0;

		rv = dvd_read(dvdid, 6, 0);
		DVD_CHECK();

		rv = dvd_read(checkBuffer, 32, 0x8001);
		DVD_CHECK();
		if (strncmp((const char *) dvdid, "COBRAM", 6) == 0) {
			multiboot();

		} else if ((strncmp((const char *) dvdid, "GGCOSD", 6) == 0)
			|| (strncmp((const char *) dvdid, "GCOPDV", 6) == 0)
			|| (strncmp((const char *) dvdid, "RGCOSD", 6) == 0)) {
			multiboot();

		} else if ((strcmp((const char *) dvdid, "") == 0)
			&& (!strncmp((const char *) checkBuffer, "\2CD001\1", 8) == 0)) {
			printf("   Media wasn't recognized!\n");
			printf
				("   Press A to try loading by MIOS without backup support...\n");
			printf("   Press B to use patched MIOS for start\n");
			printf("   Press Z to retry...\n");
			dvd_motor_off();
			while (1) {
				PAD_ScanPads();
				int buttonsDown = PAD_ButtonsDown(0);

				if (buttonsDown & PAD_BUTTON_A) {
					retry = 0;
					break;
				}
				if (buttonsDown & PAD_BUTTON_B) {
					/* Start using MIOS with backup support. */
					load_mios();
					break;
				}
				if (buttonsDown & PAD_TRIGGER_Z) {
					retry = -1;
					break;
				}
			}
		} else {
			retry = 0;
			FinalizeBoot(1);
		}
	}
}
