/*
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */
#include <stdio.h>
#include <unistd.h>
#include <ogcsys.h>
#include <gccore.h>
#include <ogc/cache.h>
#include <string.h>

#include "dvd.h"
#include "dvdreadpatch.h"
#include "dvdresetpatch.h"
#include "gameplugin_bin.h"
#include "gcplugin.h"
#include "anaconda.h"
#include "debugprintf.h"
#include "loadmios.h"

#define DVD_CHECK() \
	if (rv != 0) { \
		printf("   Failed to read DVD %d.\n", __LINE__); \
		sleep(5); \
		call_mios(); \
	}

#define DVD_BUFFER 0xC0100000
#define APPLOADER_ADDRESS 0x81200000

typedef struct {
	const char *gameid;
	u32 address1; /* Start address for patch search of second (fourth,...) DOL. */
	u32 size1; /* Size of patch. */
	u32 number1; /* Number of DOL changes. */
	u32 address2; /* Start address for patch search of first (third,...) DOL (Game Loader). */
	u32 size2; /* Size of Patch. */
	u32 number2; /* Number of DOL changes. */
} patch_entry_t;

patch_entry_t game_patches[] = {
	{
		/* Zelda Collection. */
		.gameid = "PZLJ",

		/* Patch for Game Loader. */
		.address1 = 0x81307ebc,
		.size1 = 0x100,
		.number1 = 1,

		/* Patch for final Game. */
		//.address2 = 0x800b1ca0,
		.address2 = 0x8030a4ac,
		.size2 = 0x100,
		.number2 = 1,
	}
};

void nothing()
{
}

void load_app(int reloader)
{
	gcplugin_jumptable_t *jmp1 = (gcplugin_jumptable_t *) MIOS_PLUGIN_BASE;
	gcplugin_jumptable_t *jmp2 = (gcplugin_jumptable_t *) (((u32) MIOS_PLUGIN_BASE) + sizeof(gcplugin_jumptable_t));
	int rv;
	u32 dvdid;
	char gamecode[32];

	void (*app_init)(void (*report)(const char* fmt, ...));
	int (*app_main)(void** dst, int* size, int* offset);
	void* (*app_final)();

	void (*app_entry)(void(**init)(void (*report)(const char* fmt, ...)),
		int (**main)(), void *(**final)());

	char* buffer = (char*)DVD_BUFFER;

	rv = dvd_read(buffer, 0x400, 0);
	DVD_CHECK();

	dvdid = ((u32 *) buffer)[0];
	memcpy(gamecode, buffer, sizeof(gamecode));

	memset(buffer, 0, 0x20);
	rv = dvd_read(buffer, 0x20, 0x2440);
	DVD_CHECK();

	debug_printf("Reading apploader buffer 0x%x, len 0%x, offset 0x%x\n", (void*)APPLOADER_ADDRESS, ((*(unsigned long*)(buffer + 0x14)) + 31) & ~31, 0x2460);
	rv = dvd_read((void*)APPLOADER_ADDRESS, ((*(unsigned long*)(buffer + 0x14)) + 31) & ~31, 0x2460);
	DVD_CHECK();

	app_entry = (void (*)(void(**)(void (*)(const char*, ...)), int (**)(), void *(**)()))(*(unsigned long*)(buffer + 0x10));
	app_entry(&app_init, &app_main, &app_final);
	app_init((void (*)(const char*, ...))nothing);

	/* Install DVD replacement functions. */
	memcpy(GAME_PLUGIN_BASE, gameplugin_bin, gameplugin_bin_size);

	/* Check for Reloader. */
	if (jmp1->magic == RELOAD_MAGIC) {
		int i;

		/* Reloader supported. */
		debug_printf("Reloader Supported.\n");

		if (reloader) {
			/* Try to patch every DVD transfer in the first DOL. */
			jmp2->reload_patch_start = 0;
			jmp2->reload_patch_size = 0;
			jmp2->reload_patch_number = 1;
		
			/* Try to patch every DVD transfer in the second DOL. */
			jmp1->reload_patch_start = 0;
			jmp1->reload_patch_size = 0;
			jmp1->reload_patch_number = 1;
		} else {
			for (i = 0; i < (sizeof(game_patches)/sizeof(patch_entry_t)); i++) {
				if (*((u32 *) game_patches[i].gameid) == dvdid) {
					debug_printf("Patch for game found.\n");
	
					/* Enable patching. */
					jmp2->reload_patch_start = game_patches[i].address1;
					jmp2->reload_patch_size = game_patches[i].size1;
					jmp2->reload_patch_number = game_patches[i].number1;
		
					jmp1->reload_patch_start = game_patches[i].address2;
					jmp1->reload_patch_size = game_patches[i].size2;
					jmp1->reload_patch_number = game_patches[i].number2;
				}
			}
		}
	} else {
		debug_printf("Reloader Supported.\n");
	}

	DCFlushRange(GAME_PLUGIN_BASE, gameplugin_bin_size);
	ICInvalidateRange(GAME_PLUGIN_BASE, gameplugin_bin_size);

	while(1)
	{
		void* dst = 0;
		int len = 0,offset = 0;
		int res = app_main(&dst, &len, &offset);

		if (!res)
			break;
		debug_printf("Reading section buffer 0x%x, len 0%x, offset 0x%x\n", dst, len, offset);
		rv = dvd_read(dst, len, offset);
		DVD_CHECK();
		dvd_patchreset(dst, len, GAME_PLUGIN_BASE);
		dvd_patchreadid(dst, len, GAME_PLUGIN_BASE);
		dvd_patchread(dst, len, GAME_PLUGIN_BASE, 0, gamecode);

		DCFlushRange(dst, len);
		ICInvalidateRange(dst, len);
	}
	DCFlushRange(GAME_PLUGIN_BASE, gameplugin_bin_size);
	ICInvalidateRange(GAME_PLUGIN_BASE, gameplugin_bin_size);

	void (*entrypoint)() = app_final();
	entrypoint();
}

