#include <stdio.h>
#include <stdlib.h>
#include <ogc/cache.h>
#include <string.h>

#include "loadapp.h"
#include "dvdreadpatch.h"
#include "dvdresetpatch.h"
#if !(defined(RELOAD_SUPPORT) && defined(GEKKO_DEBUG)) /* Module can't be generated if too much is compiled in. */
#include "miosplugin_bin.h"
#endif
#include "gcplugin.h"
#include "loadmios.h"
#include "debugprintf.h"

/** Start address of MIOS v5. */
#define MIOS_BASE ((void *) 0x812fffe0)
/** Size of MIOS v5. */
#define MIOS_SIZE 0x0002659c
/** Entry point for MIOS v5. */
#define MIOS_ENTRY 0x81300200
/** Entry point variable. */
#define MIOS_PATCH_ADDR ((void *) 0x800037fc)

void call_mios(void)
{
	unsigned long *entry = MIOS_PATCH_ADDR;

	/* Restore patched entry point. */
	*entry = MIOS_ENTRY;
	DCFlushRange(entry, 32);

	/* Simulate boot. */
	__asm__(
		"bl DCDisable\n"
		"bl ICDisable\n"
		"li %r3, 0\n"
		"mtsrr1 %r3\n"
		"li %r4, 0x3400\n"
		"mtsrr0 %r4\n"
		"rfi\n"
	);
}

void load_mios(void)
{
#if !(defined(RELOAD_SUPPORT) && defined(GEKKO_DEBUG)) /* Module can't be generated if too much is compiled in. */
	/* Start game using MIOS. */
	void *dst = MIOS_BASE;
	unsigned long len = MIOS_SIZE;
	gcplugin_jumptable_t *jmp = (gcplugin_jumptable_t *) MIOS_PLUGIN_BASE;

	/* Install DVD replacement functions. */
	memcpy(MIOS_PLUGIN_BASE, miosplugin_bin, miosplugin_bin_size);
	debug_printf("magic 0x%08x\n", jmp->magic);
	if (jmp->magic == RELOAD_MAGIC) {
		/* Reloader supported. */
		debug_printf("Reloader Supported.\n");

		/* First table is used for MIOS. */

		/* Second table is used for games. */
		jmp = (gcplugin_jumptable_t *) (((u32) MIOS_PLUGIN_BASE) + sizeof(gcplugin_jumptable_t));

		/* Enable patching. */
		jmp->reload_patch_start = 0;
		jmp->reload_patch_size = 0;
		jmp->reload_patch_number = 1;
	}
	DCFlushRange(MIOS_PLUGIN_BASE, miosplugin_bin_size);
	ICInvalidateRange(MIOS_PLUGIN_BASE, miosplugin_bin_size);

	dvd_patchreset(dst, len, MIOS_PLUGIN_BASE);
	dvd_patchreadid(dst, len, MIOS_PLUGIN_BASE);
	dvd_patchread(dst, len, MIOS_PLUGIN_BASE, 1, NULL);
	DCFlushRange(dst, len);
	ICInvalidateRange(dst, len);

	DCFlushRange(MIOS_PLUGIN_BASE, miosplugin_bin_size);
	ICInvalidateRange(MIOS_PLUGIN_BASE, miosplugin_bin_size);
#endif

	/* Jump to MIOS entry. */
	call_mios();
}
