/* Copyright 2009 WiiGator. */
#include "types.h"

#include "debugprintf.h"
#include "dvdreadpatch.h"
#include "writebranch.h"
#include "codecmp.h"
#include "cache.h"

/** Maximum size of the patch (needed to cache flush). */
#define MAX_PATCH_SIZE 0x100

#define NORMAL_DOL_ADDRESS 0x80140000
#define NORMAL_DOL_START 0x80003100

/** Original DVD read function begins with the following bytes: */
static const u32 _Read_original[] = {
	0x7C0802A6,                             
	0x90010004,                             
	0x38000000,                             
	0x9421FFD8,                             
	0x93E10024,                             
	0x93C10020,                             
	0x3BC50000	                             
};

static const u32 _Read_original2[] = {
	0x7c0802a6,
	0x90010004,
	0x9421ffe0,
	0x93e1001c,
	0x90610008,
	0x7c9f2378,
	0x90a10010,
	0x90c10014,
	0x80010014,
	0x900da720,
	0x38000000,
	0x900da718,
	0x38000001,
	0x900da758,
};

static const u32 _Read_original3[] = {
	0x7c0802a6,
	0x90010004,
	0x38000001,
	0x9421ffd8,
	0x93e10024,
	0x93c10020,
	0x3bc50000,
	0x93a1001c,
	0x3ba40000,
	0x93810018,
	0x3b830000,
	0x90cd16bc,
	0x3cc0804d,
	0x3be6bdc0,
	0x900d16f8
};

#if 0 /* Should be only for MIOS required, but MIOS is already patched, when this is running.  */
static const u32 _Read_originalMIOSv5[] = {
	0x9421fff0,
	0x7c0802a6,
	0x38e00000,
	0x90010014,
	0x93e1000c,
	0x93c10008,
	0x800d8274,
	0x90cd8230,
	0x2c000000,
	0x90ed8208,
	0x40820008,
	0x3ca50009,
	0x3ce0cc00,
	0x3c00a800,
	0x90076008
};
#endif

#ifdef GEKKO_DEBUG
static const u32 stream_write_orignal[] = {
	0x7c0802a6,
	0x90010004,
	0x9421ffe0,
	0x93e1001c,
	0x3be50000,
	0x93c10018,
	0x3bc30000,
	0x38640000,
	0x38de0000,
	0x38800001
};


static const u32 stream_write_orignal2[] = {
	0x9421fff0,
	0x7c0802a6,
	0x90010014,
	0x93e1000c,
	0x7cbf2b78,
	0x93c10008,
	0x7c7e1b78,
	0x7c832378,
	0x38800001,
	0x7fc6f378
};

static const u32 stream_write_orignal3[] = {
	0x9421ffe0,
	0x7c0802a6,
	0x90010024,
	0x93e1001c,
	0x93c10018,
	0x7cbe2b78,
	0x90810008,
	0x38800001,
	0x93a10014,
	0x7c7d1b78,
	0x80610008,
	0x7fa6eb78
};
#endif

/** Patch DVD read function. The patch should be applied only one time. */
void dvd_patchread(void *addr, u32 len, gcplugin_jumptable_t *jmp) {
	void *addr_start = addr;
	void *addr_end = addr+len;
	int inRange = 0;
	static int lastJmpNr = 0;

	if (lastJmpNr != jmpNr) {
		if (lastJmpNr == 0) {
			if (jumptable2.reload_patch_number > 0) {
				jumptable2.reload_patch_number--;
			}
		} else {
			if (jumptable1.reload_patch_number > 0) {
				jumptable1.reload_patch_number--;
			}
		}
		lastJmpNr = jmpNr;
	}

	/* Game will be slow if everything is checked, so reduce area checked: */
	if (jmp->reload_patch_number == 0) {
		/* Maximum number of reloads reached. */
		return;
	}

	if (jmp->reload_patch_start != 0) {
		/* Enabled special range. */
		if ((jmp->reload_patch_start >= ((u32) addr))
			&& (jmp->reload_patch_start < (((u32) addr) + len))) {
			addr = (void *) jmp->reload_patch_start;
			len = jmp->reload_patch_size;
		} else {
			/* First memory area is normally used for DOLs. */
			if (!((NORMAL_DOL_ADDRESS >= ((u32) addr))
				&& (NORMAL_DOL_START < (((u32) addr) + len)))) {
				return;
			}
		}
	}

	/* Set DVD base offset for DVD read commands. */
	jmp->base_offset = base_offset1;

	while(addr_start<addr_end) {
		if (codecmp(addr_start,_Read_original,sizeof(_Read_original))==0)  {
			debug_printf("Patching DVD read at 0x%x\n", (u32) addr_start);

			install_replacement((void*)addr_start, jmp->dvd_read_replacement, jmp->dvd_read_orig);

			// Change DVD command from 0xA8 to 0xD0:
			*(unsigned long*)(addr_start +0x5C) = 0x3C60D000;
			// Shift (>>11) the size and put in r3:
			*(unsigned long*)(addr_start +0x64) = 0x57A3AAFE;
			// Don't shift (>>2) the offset, because this is now LBA:
			*(unsigned long*)(addr_start +0x6C) = 0x93C4000C;
			// Store shifted size value:
			*(unsigned long*)(addr_start +0x74) = 0x90640010;
		}

		if (codecmp(addr_start,_Read_original2,sizeof(_Read_original2))==0)  {
			debug_printf("Patching DVD read 2 at 0x%x\n", (u32) addr_start);

			install_replacement((void*)addr_start, jmp->dvd_read_replacement, jmp->dvd_read_orig);

			// Change DVD command from 0xA8 to 0xD0:
			*(unsigned long*)(addr_start +0x44) = 0x3C00D000;
			// Don't shift (>>2) the offset, because this is now LBA:
			// Shift (>>11) the size and put in r4:
			*(unsigned long*)(addr_start +0x54) = 0x57E4AAFE;
			// Store shifted size value:
			*(unsigned long*)(addr_start +0x64) = 0x90836010;
		}

		if (codecmp(addr_start,_Read_original3,sizeof(_Read_original3))==0)  {
			debug_printf("Patching DVD read 3 at 0x%x\n", (u32) addr_start);

			install_replacement((void*)addr_start, jmp->dvd_read_replacement, jmp->dvd_read_orig);

			// Change DVD command from 0xA8 to 0xD0:
			*(unsigned long*)(addr_start +0x54) = 0x3C60D000;
			// Shift (>>11) the size and put in r3:
			*(unsigned long*)(addr_start +0x5c) = 0x57A3AAFE;
			// Don't shift (>>2) the offset, because this is now LBA:
			*(unsigned long*)(addr_start +0x64) = 0x93C4000C;
			// Store shifted size value:
			*(unsigned long*)(addr_start +0x6c) = 0x90640010;
		}

#if 0 /* Should be only for MIOS required, but MIOS is already patched, when this is running.  */
		if (codecmp(addr_start,_Read_originalMIOSv5,sizeof(_Read_originalMIOSv5))==0)  {
			debug_printf("Patching DVD read MIOSv5 at 0x%x\n", (u32) addr_start);

			install_replacement((void*)addr_start, jmp->dvd_read_replacement, jmp->dvd_read_orig);

			// Shift (>>11) the size and put in r30:
			*(unsigned long*)(addr_start +0x28) = 0x549EAAFE;
			// Remove offset + 9
			*(unsigned long*)(addr_start +0x2c) = 0x60000000;
			// Change DVD command from 0xA8 to 0xD0:
			*(unsigned long*)(addr_start +0x34) = 0x3C00D000;
			// Store shifted size value:
			*(unsigned long*)(addr_start +0x48) = 0x93C76010;
		}
#endif

#ifdef GEKKO_DEBUG
		/* Patch debug printf. */
		if ((codecmp(addr_start, stream_write_orignal, sizeof(stream_write_orignal))==0)
			|| (codecmp(addr_start, stream_write_orignal2, sizeof(stream_write_orignal2))==0)
			|| (codecmp(addr_start, stream_write_orignal3, sizeof(stream_write_orignal3))==0)) {
			if (jmp->stream_write_replacement != 0) {
				debug_printf("Patching debug write at 0x%x\n", (u32) addr_start);

				writebranch((void*)addr_start, jmp->stream_write_replacement);

				DCFlushRange(addr_start, MAX_PATCH_SIZE);
			}
		}
#endif
		addr_start += 4;
	}
}
